unit ucalibrarconoensamble;

{$mode delphi}

interface

uses
  uauxiliares,
  Math,
  uopencalc,
  fpsTypes,
  umodelosintcegh,
  upronostico,
  xmatdefs,
  ufechas,
  ufxgx,
  uresfxgx,
  ucalibrarconopronosticos,
  useriestemporales,
  ucosaConNombre,
  ucosa,
  ucosanubeseable,
  uInicioYFinal,
  Classes,
  SysUtils,
  matreal,
  uFuenteSintetizador,
  uestadosinuy,
  usalasdejuego,
  uproblema,
  uDataSetGenerico,
  uoddface_roscon_cfg,
  udbconpg, urosx;

{GFlieller 29-11-2018
Ajustar el cono de un CEGH a ensamble de pronosticos
1- Buscar las matrices A y B en el CEGH. Calcular la matriz C=(B B^T)^-1
2- Transformar el ensamble a espacio gaussiano
3- Calcular los sesgos
4- Calcular los desvios
5- Calcular la matriz de covarianzas
6- Calcular la matriz de ajuste con el paso anterior "Z"
7- Calcular la matriz M(covar,Z)
8- Resolver MX = "1/X"
9- Volver al paso 6

}

type

    { TcalibrarConoEnsamble }

    TcalibrarConoEnsamble = class
      A : TMatR;
      B : TMatR;
      C : TMatR;

      fcegh : TFuenteSintetizadorCEGH;

      EnsamblesGauss : array of TMatR;
      Xg : array of array of TMatR;
      Xesp : array of TMatR;

      Ncanales : integer; // Series [Bonete,Palmar,Salto]
      Nmiembros : integer;
      Npasos : integer;

      M : array of TMatR;
      Z : array of TMatR;
      S : array of TMatR;
      betas : array of TVectR;
      Beta : TMatR;
      Lambda : TMatR;
      Lambdas : array of TMatR;
      factores : array of TDAofNReal;
      sesgos: array of TDAofNReal;
      des : array of array of TMatR;
      covar : array of TMatR;

      aPronostico :array of TPronostico;
      aPronosticos: TPronosticos;

      constructor Create(acegh: TFuenteSintetizadorCEGH; aEnsambles: array of TMatR);

      function calibrarConoEnsamble() : boolean;

      procedure cargarDimensiones(aEnsambles: array of TMatR);
      // carga las dimensiones: Ncanales, Nmiembros, Nensambles

      procedure cargarEnsamblesGausiano(aEnsambles: array of TMatR);
      // carga todos los ensambles a la clase

      procedure calcularGuias;
      // Calcula las guias en mundo gaussiano y real

      procedure calcularSesgos;
      // Calcula los sesgos S_k

      procedure calcularCovar;
      // Con los desvios calculados y cargados es des[][] calcula las matrices de covarianza

      function calcularZ(Z0 : TMatR; kpaso : integer) : TMatR;
      // Calcula la matriz que en la doc se denota como Z. Corresponde a un
      // ajuste de la varianza entre pasos

      procedure cargarFiltros;
      // Carga los filtros A y B del modelo. Ademas calcula la matriz C = ( B x B^{T} )^{-1}

      procedure armarProblema(Z : TMatR; kpaso : integer);
      // Arma la matriz M del problema M * X = 1/X

      procedure ImprimirLambdas;
      // Guarda las matrices de lambdas

      procedure guardarSala(sala: TSalaDeJuego);

      function TransformarEnsamble(aModelo:TModeloCEGH; aEnsambles:array of TMatR; hcanal: integer) : TMatR;
      // Trasforma el ensamble a espacio gaussiano

      private procedure ArmarResultados(aModelo: TModeloCEGH);
      // Carga al modelo CEGH los valores de guia, guia gaussina, sesgos,
      // factores, NPAC, NPSA, NPCC

      procedure calcularDesvios;
      // En Tpronosticos, que es la clase que utiliza el Sintetizador, la guia,
      // los sesgos y los factores son TDAofNReal, pero para manipular estos
      // valores en la unidad se usan arrays of TMatR. Acá lo que se hace es
      // pasar los valores que se tienen al formato que requiere el Sintetizador.

      procedure free;

end;

type

  { T_fobjetivoCEGH }

  T_fobjetivoCEGH = class(Tfx)


  Error: NReal;
  M : TMatR;
  FX : TVectR;
  Ncanales : integer;

  constructor Create(aM: TMatR);
  // Crea la funcion objetivo usando la matriz M del problema

  function f(const X: TVectR): NReal; override;
  // Calculo de costo.

  procedure free; override;

  procedure acum_g(var grad: TVectR; const X: TVectR); override;
  // calculo de gradiente. El gradiente depende de FX, X y M
  end;

type

    { TProblemaResolverFactores }
    TProblemaResolverFactores = class(TProblema_m01)

    M: TMatR;                 // Matriz del problema
    Betas: TVectR;
    Lambdas: TMatR;
    Ncanales:integer;

    constructor Create(aM:TMatR);
    // Crea el problema M * X = 1/X

    procedure free;

    function resolverFactores(var factores:TMatR): boolean;
    // Resuelve el problema M * X = 1/X

    end;



type

    { TdatosHistoricosVates }

    TdatosHistoricosVates = class

    // Cada Serie es un pronostico de 16 dias producido por el vates
    //      Se les carga, nombre, fecha inicio y dt entre muestras. Estas series
    //      son de paso horario
    pronosticos: array of TSeriesDeDatos;
    pronosticosPasoDiario: array of TSeriesDeDatos;

    // Se arma la serie de valores que tiene la tira de valores iniciales para
    //      cada dia. Los nombres son Bonete, Baygorria, Palmar y SG; los datos
    //      resampleados a diario de pronosBon, pronosBay, pronosPal y pronosSG
    //      se cargan la serie valores iniciales.
    EstadosIniciales: TSeriesDeDatos;
    NombresSeries: array [0..4] of string;
    // Nombre de cada serie en la base de datos


    sqlbase1 : string;
    sqlbase2  : string;
    FechaIni : TDateTime;
    FechaFin : TDateTime;
    Fecha    : TDateTime;
    dias     : integer;
    horas    : integer;
    series   : integer;

    constructor Create;
    procedure descargarPronos;
    procedure armarIniciales;
    procedure free;


    end;


//  Funciones sueltas
procedure WriteToArchiODSResultdosSim(DatosEnsamble,DatosSimCEGH: array of TMatR; archi : string;formato:TsSpreadsheetFormat=sfOpenDocument);

procedure WriteToArchiXLTResultdosSim(DatosEnsamble,DatosSimCEGH: array of TMatR; archi : string);

procedure bajarEstadosInicialesHistIMFIA( var estado : TEstado_SIN_UY; dbpqPronos: TDBPQCon; Ndias: integer);

procedure cargarEnsamblesLocales(var SeriesDeEnsambles : array of TMatR;archidir: string;Nseries: integer);
          // Carga los ramilletes de pronosticos del IMIFIA a partir de archivo .xlt
          // guardados localmente. Se usa para debuguear el programa.
procedure cargarEnsamblesDesdeEstado(var SeriesDeEnsambles : array of TMatR;estado:TEstado_SIN_UY;Nseries: integer);
          // Carga los ramilletes de pronosticos del IMFIA a partir del estado. En el
          // estado debe estar precargados los ramilletes a traves de consulta sql y
          // corrida del modelo del IMFIA.
procedure graficarResultados;

procedure graficarResultados_;

procedure bajarHistoricos;

implementation

procedure WriteToArchiXLTResultdosSim(DatosEnsamble,
  DatosSimCEGH: array of TMatR; archi: string);
var
  SeriesPE_CEGH, SeriesPE_IMFIA : array of TMatR;               // Series de probabilidad de excedencia
  PE5_CEGH,PE10_CEGH,PE20_CEGH,PE50_CEGH,PE80_CEGH,PE90_CEGH,PE95_CEGH, PE5_IMFIA,PE10_IMFIA,PE20_IMFIA,PE50_IMFIA,PE80_IMFIA,PE90_IMFIA,PE95_IMFIA : array of TVectR;   // Cortes de probabilidad
  kSerie: Integer;
begin
  // 5.1  Paso los ensambles a .xlt
  DatosEnsamble[0].WriteArchiXLT('/home/'+GetEnvironmentVariable('USER')+'/prono_hidro_imfia/Resultados/Bonete/Ensamble/'+archi+'.xlt');
  if length(DatosEnsamble)>1 then
    DatosEnsamble[1].WriteArchiXLT('/home/'+GetEnvironmentVariable('USER')+'/prono_hidro_imfia/Resultados/Palmar/Ensamble/'+archi+'.xlt');
  if length(DatosEnsamble)>2 then
    DatosEnsamble[2].WriteArchiXLT('/home/'+GetEnvironmentVariable('USER')+'/prono_hidro_imfia/Resultados/SG/Ensamble/'+archi+'.xlt');

  // 5.2  Paso los ramilletes de sinteticas a .xlt
  DatosSimCEGH[0].WriteArchiXLT('/home/'+GetEnvironmentVariable('USER')+'/prono_hidro_imfia/Resultados/Bonete/Sinteticas/'+archi+'.xlt');
  if length(DatosSimCEGH)>1 then
    DatosSimCEGH[1].WriteArchiXLT('/home/'+GetEnvironmentVariable('USER')+'/prono_hidro_imfia/Resultados/Palmar/Sinteticas/'+archi+'.xlt');
  if length(DatosSimCEGH)>2 then
    DatosSimCEGH[2].WriteArchiXLT('/home/'+GetEnvironmentVariable('USER')+'/prono_hidro_imfia/Resultados/SG/Sinteticas/'+archi+'.xlt');

  for kSerie := 0 to (length(DatosSimCEGH)-1) do
  begin
  // 5.3  Ordeno las series para sacar cada corte de probabilidad
  DatosSimCEGH[kSerie].Sort();
  // 5.4  Cargo cada corte de porbabilidad a la serie correspondiente
  PE5_CEGH[kSerie]:= DatosSimCEGH[kSerie].Columna(round(0.05*DatosSimCEGH[kSerie].nc));
  PE10_CEGH[kSerie]:=DatosSimCEGH[kSerie].Columna(round(0.10*DatosSimCEGH[kSerie].nc));
  PE20_CEGH[kSerie]:=DatosSimCEGH[kSerie].Columna(round(0.20*DatosSimCEGH[kSerie].nc));
  PE50_CEGH[kSerie]:=DatosSimCEGH[kSerie].Columna(round(0.50*DatosSimCEGH[kSerie].nc));
  PE80_CEGH[kSerie]:=DatosSimCEGH[kSerie].Columna(round(0.80*DatosSimCEGH[kSerie].nc));
  PE90_CEGH[kSerie]:=DatosSimCEGH[kSerie].Columna(round(0.90*DatosSimCEGH[kSerie].nc));
  PE95_CEGH[kSerie]:=DatosSimCEGH[kSerie].Columna(round(0.95*DatosSimCEGH[kSerie].nc));
  // 5.5  Creo matrices con todos los cortes de probabilidad para cada canal
  SeriesPE_CEGH[kSerie]:=TMatR.Create_Init(336,7);
  SeriesPE_CEGH[kSerie].pon_columna(1,1,PE5_CEGH[kSerie].toTDAOfNReal());
  SeriesPE_CEGH[kSerie].pon_columna(1,2,PE10_CEGH[kSerie].toTDAOfNReal());
  SeriesPE_CEGH[kSerie].pon_columna(1,3,PE20_CEGH[kSerie].toTDAOfNReal());
  SeriesPE_CEGH[kSerie].pon_columna(1,4,PE50_CEGH[kSerie].toTDAOfNReal());
  SeriesPE_CEGH[kSerie].pon_columna(1,5,PE80_CEGH[kSerie].toTDAOfNReal());
  SeriesPE_CEGH[kSerie].pon_columna(1,6,PE90_CEGH[kSerie].toTDAOfNReal());
  SeriesPE_CEGH[kSerie].pon_columna(1,7,PE95_CEGH[kSerie].toTDAOfNReal());
  end;

  SeriesPE_CEGH[0].WriteArchiXLT('/home/'+GetEnvironmentVariable('USER')+'/prono_hidro_imfia/Resultados/Bonete/Sinteticas_PE/'+archi+'.xlt');
  if length(DatosEnsamble)>1 then
    SeriesPE_CEGH[1].WriteArchiXLT('/home/'+GetEnvironmentVariable('USER')+'/prono_hidro_imfia/Resultados/Palmar/Sinteticas_PE/'+archi+'.xlt');
  if length(DatosEnsamble)>2 then
    SeriesPE_CEGH[2].WriteArchiXLT('/home/'+GetEnvironmentVariable('USER')+'/prono_hidro_imfia/Resultados/SG/Sinteticas_PE/'+archi+'.xlt');


  for kSerie := 0 to (length(DatosEnsamble)-1) do
    begin
    // 5.5  Ordeno las series para sacar cada corte de probabilidad
    DatosEnsamble[kSerie].Sort();
    // 5.6  Cargo cada corte de porbabilidad a la serie correspondiente
    PE5_IMFIA[kSerie] := DatosEnsamble[kSerie].Columna(round(0.05*DatosEnsamble[kSerie].nc));
    PE10_IMFIA[kSerie]:=DatosEnsamble[kSerie].Columna(round(0.10*DatosEnsamble[kSerie].nc));
    PE20_IMFIA[kSerie]:=DatosEnsamble[kSerie].Columna(round(0.20*DatosEnsamble[kSerie].nc));
    PE50_IMFIA[kSerie]:=DatosEnsamble[kSerie].Columna(round(0.50*DatosEnsamble[kSerie].nc));
    PE80_IMFIA[kSerie]:=DatosEnsamble[kSerie].Columna(round(0.80*DatosEnsamble[kSerie].nc));
    PE90_IMFIA[kSerie]:=DatosEnsamble[kSerie].Columna(round(0.90*DatosEnsamble[kSerie].nc));
    PE95_IMFIA[kSerie]:=DatosEnsamble[kSerie].Columna(round(0.95*DatosEnsamble[kSerie].nc));
    // 5.7  Creo matrices con todos los cortes de probabilidad para cada canal
    SeriesPE_IMFIA[kSerie]:=TMatR.Create_Init(336,7);
    SeriesPE_IMFIA[kSerie].pon_columna(1,1,PE5_IMFIA[kSerie].toTDAOfNReal());
    SeriesPE_IMFIA[kSerie].pon_columna(1,2,PE10_IMFIA[kSerie].toTDAOfNReal());
    SeriesPE_IMFIA[kSerie].pon_columna(1,3,PE20_IMFIA[kSerie].toTDAOfNReal());
    SeriesPE_IMFIA[kSerie].pon_columna(1,4,PE50_IMFIA[kSerie].toTDAOfNReal());
    SeriesPE_IMFIA[kSerie].pon_columna(1,5,PE80_IMFIA[kSerie].toTDAOfNReal());
    SeriesPE_IMFIA[kSerie].pon_columna(1,6,PE90_IMFIA[kSerie].toTDAOfNReal());
    SeriesPE_IMFIA[kSerie].pon_columna(1,7,PE95_IMFIA[kSerie].toTDAOfNReal());
    end;


  SeriesPE_IMFIA[0].WriteArchiXLT('/home/'+GetEnvironmentVariable('USER')+'/prono_hidro_imfia/Resultados/Bonete/Ensambles_PE/'+archi+'.xlt');
  if length(SeriesPE_IMFIA)>1 then
    SeriesPE_IMFIA[1].WriteArchiXLT('/home/'+GetEnvironmentVariable('USER')+'/prono_hidro_imfia/Resultados/Palmar/Ensambles_PE/'+archi+'.xlt');
  if length(SeriesPE_IMFIA)>2 then
    SeriesPE_IMFIA[2].WriteArchiXLT('/home/'+GetEnvironmentVariable('USER')+'/prono_hidro_imfia/Resultados/SG/Ensambles_PE/'+archi+'.xlt');

  // 5.8  Libero memoria
  for kSerie := 0 to (length(DatosSimCEGH)-1) do
  begin
    PE5_CEGH[kserie].free;
    PE10_CEGH[kserie].free;
    PE20_CEGH[kserie].free;
    PE50_CEGH[kserie].free;
    PE80_CEGH[kserie].free;
    PE90_CEGH[kserie].free;
    PE95_CEGH[kserie].free;
    SeriesPE_CEGH[kSerie].free;
  end;
  for kSerie := 0 to (length(DatosEnsamble)-1) do
  begin
    PE5_IMFIA[kserie].free;
    PE10_IMFIA[kserie].free;
    PE20_IMFIA[kserie].free;
    PE50_IMFIA[kserie].free;
    PE80_IMFIA[kserie].free;
    PE90_IMFIA[kserie].free;
    PE95_IMFIA[kserie].free;
    SeriesPE_IMFIA[kSerie].free;
  end;

  SetLength(PE5_CEGH ,0);
  SetLength(PE10_CEGH,0);
  SetLength(PE20_CEGH,0);
  SetLength(PE50_CEGH,0);
  SetLength(PE80_CEGH,0);
  SetLength(PE90_CEGH,0);
  SetLength(PE95_CEGH,0);
  SetLength(SeriesPE_CEGH,0);
  SetLength(PE5_IMFIA,0);
  SetLength(PE10_IMFIA,0);
  SetLength(PE20_IMFIA,0);
  SetLength(PE50_IMFIA,0);
  SetLength(PE80_IMFIA,0);
  SetLength(PE90_IMFIA,0);
  SetLength(PE95_IMFIA,0);
  SetLength(SeriesPE_IMFIA,0);



end;

procedure WriteToArchiODSResultdosSim(DatosEnsamble,DatosSimCEGH: array of TMatR; archi: string; formato:TsSpreadsheetFormat=sfOpenDocument);
// 5. Guardo en un ods los conos del ensamble y los conos creados con el CEGH
//    como cortes de probabilidad
var
ods: TLibroOpenCalc;
SeriesPE_CEGH, SeriesPE_IMFIA : array of TMatR;               // Series de probabilidad de excedencia
PE5_CEGH,PE10_CEGH,PE20_CEGH,PE50_CEGH,PE80_CEGH,PE90_CEGH,PE95_CEGH, PE5_IMFIA,PE10_IMFIA,PE20_IMFIA,PE50_IMFIA,PE80_IMFIA,PE90_IMFIA,PE95_IMFIA : array of TVectR;   // Cortes de probabilidad
kSerie: Integer;

begin
// 5.1  Creo el libro ods que va contener todas las hojas
if (formato = TsSpreadsheetFormat.sfOpenDocument) or (formato = TsSpreadsheetFormat.sfOOXML) then
   ods := TLibroOpenCalc.Create(TRUE,archi,False,formato)
else
   writeln('El formato de los resultados debe ser .xlsx o .ods');

// 5.2  Agrego cada hoja de acuerdo a la cantidad de datos que tenga
ods.agregoHoja('Bonete_IMFIA');
if length(DatosEnsamble)>1 then
   ods.agregoHoja('Palmar_IMFIA');
if length(DatosEnsamble)>2 then
   ods.agregoHoja('SG_IMFIA');

// 5.3  Escribo el contenido de cada serie
ods.Write_MatR(DatosEnsamble[0],'Bonete_IMFIA',1,1);
if length(DatosEnsamble)>1 then
   ods.Write_MatR(DatosEnsamble[1],'Palmar_IMFIA',1,1);
if length(DatosEnsamble)>2 then
   ods.Write_MatR(DatosEnsamble[2],'SG_IMFIA',1,1);

// 5.2  Agrego cada hoja de acuerdo a la cantidad de datos que tenga
ods.agregoHoja('Bonete_CEGH');
if length(DatosSimCEGH)>1 then
   ods.agregoHoja('Palmar_CEGH');
if length(DatosSimCEGH)>2 then
   ods.agregoHoja('SG_CEGH');

// 5.3  Escribo el contenido de cada serie
ods.Write_MatR(DatosSimCEGH[0],'Bonete_CEGH',1,1);
if length(DatosSimCEGH)>1 then
   ods.Write_MatR(DatosSimCEGH[1],'Palmar_CEGH',1,1);
if length(DatosSimCEGH)>2 then
   ods.Write_MatR(DatosSimCEGH[2],'SG_CEGH',1,1);

/////////////////////////////////////////////////////////////////

// 5.4  Fijo la cantidad de canales de cada corte de probabilidad
setlength(PE5_CEGH,length(DatosSimCEGH));
setlength(PE10_CEGH,length(DatosSimCEGH));
setlength(PE20_CEGH,length(DatosSimCEGH));
setlength(PE50_CEGH,length(DatosSimCEGH));
setlength(PE80_CEGH,length(DatosSimCEGH));
setlength(PE90_CEGH,length(DatosSimCEGH));
setlength(PE95_CEGH,length(DatosSimCEGH));
setlength(SeriesPE_CEGH,length(DatosSimCEGH));

for kSerie := 0 to (length(DatosSimCEGH)-1) do
  begin
  // 5.5  Ordeno las series para sacar cada corte de probabilidad
  DatosSimCEGH[kSerie].Sort();
  // 5.6  Cargo cada corte de porbabilidad a la serie correspondiente
  PE5_CEGH[kSerie]:= DatosSimCEGH[kSerie].Columna(round(0.05*DatosSimCEGH[kSerie].nc));
  PE10_CEGH[kSerie]:=DatosSimCEGH[kSerie].Columna(round(0.10*DatosSimCEGH[kSerie].nc));
  PE20_CEGH[kSerie]:=DatosSimCEGH[kSerie].Columna(round(0.20*DatosSimCEGH[kSerie].nc));
  PE50_CEGH[kSerie]:=DatosSimCEGH[kSerie].Columna(round(0.50*DatosSimCEGH[kSerie].nc));
  PE80_CEGH[kSerie]:=DatosSimCEGH[kSerie].Columna(round(0.80*DatosSimCEGH[kSerie].nc));
  PE90_CEGH[kSerie]:=DatosSimCEGH[kSerie].Columna(round(0.90*DatosSimCEGH[kSerie].nc));
  PE95_CEGH[kSerie]:=DatosSimCEGH[kSerie].Columna(round(0.95*DatosSimCEGH[kSerie].nc));
  // 5.7  Creo matrices con todos los cortes de probabilidad para cada canal
  SeriesPE_CEGH[kSerie]:=TMatR.Create_Init(336,7);
  SeriesPE_CEGH[kSerie].pon_columna(1,1,PE5_CEGH[kSerie].toTDAOfNReal());
  SeriesPE_CEGH[kSerie].pon_columna(1,2,PE10_CEGH[kSerie].toTDAOfNReal());
  SeriesPE_CEGH[kSerie].pon_columna(1,3,PE20_CEGH[kSerie].toTDAOfNReal());
  SeriesPE_CEGH[kSerie].pon_columna(1,4,PE50_CEGH[kSerie].toTDAOfNReal());
  SeriesPE_CEGH[kSerie].pon_columna(1,5,PE80_CEGH[kSerie].toTDAOfNReal());
  SeriesPE_CEGH[kSerie].pon_columna(1,6,PE90_CEGH[kSerie].toTDAOfNReal());
  SeriesPE_CEGH[kSerie].pon_columna(1,7,PE95_CEGH[kSerie].toTDAOfNReal());
  end;

// 5.8  Creo las hojas de los cortes de probabilidad
ods.agregoHoja('Bonete_PE_CEGH');
if length(SeriesPE_CEGH)>1 then
   ods.agregoHoja('Palmar_PE_CEGH');
if length(SeriesPE_CEGH)>2 then
   ods.agregoHoja('SG_PE_CEGH');
if length(SeriesPE_CEGH)>3 then
   ods.agregoHoja('in34_PE_CEGH');
if length(SeriesPE_CEGH)>4 then
   ods.agregoHoja('cmo1_PE_CEGH');
if length(SeriesPE_CEGH)>5 then
   ods.agregoHoja('cmo3_PE_CEGH');

// 5.9  Escribo el contenido de cada hoja
//ods.ActivoHoja('Bonete_SimCEGH');
ods.go('Bonete_PE_CEGH',1,1);
ods.write('PE5');
ods.write('PE10');
ods.write('PE20');
ods.write('PE50');
ods.write('PE80');
ods.write('PE90');
ods.write('PE95');
ods.Write_MatR(SeriesPE_CEGH[0],'Bonete_PE_CEGH',2,1);
if length(SeriesPE_CEGH)>1 then
   begin
   ods.go('Palmar_PE_CEGH',1,1);
   ods.write('PE5');
   ods.write('PE10');
   ods.write('PE20');
   ods.write('PE50');
   ods.write('PE80');
   ods.write('PE90');
   ods.write('PE95');
   ods.Write_MatR(SeriesPE_CEGH[1],'Palmar_PE_CEGH',2,1);
   end;
if length(SeriesPE_CEGH)>2 then
   begin
   ods.go('SG_PE_CEGH',1,1);
   ods.write('PE5');
   ods.write('PE10');
   ods.write('PE20');
   ods.write('PE50');
   ods.write('PE80');
   ods.write('PE90');
   ods.write('PE95');
   ods.Write_MatR(SeriesPE_CEGH[2],'SG_PE_CEGH',2,1);
   end;
if length(SeriesPE_CEGH)>3 then
   begin
   ods.go('in34_PE_CEGH',1,1);
   ods.write('PE5');
   ods.write('PE10');
   ods.write('PE20');
   ods.write('PE50');
   ods.write('PE80');
   ods.write('PE90');
   ods.write('PE95');
   ods.Write_MatR(SeriesPE_CEGH[3],'in34_PE_CEGH',2,1);
   end;
if length(SeriesPE_CEGH)>4 then
   begin
   ods.go('cmo1_PE_CEGH',1,1);
   ods.write('PE5');
   ods.write('PE10');
   ods.write('PE20');
   ods.write('PE50');
   ods.write('PE80');
   ods.write('PE90');
   ods.write('PE95');
   ods.Write_MatR(SeriesPE_CEGH[4],'cmo1_PE_CEGH',2,1);
   end;
if length(SeriesPE_CEGH)>5 then
   begin
   ods.go('cmo3_PE_CEGH',1,1);
   ods.write('PE5');
   ods.write('PE10');
   ods.write('PE20');
   ods.write('PE50');
   ods.write('PE80');
   ods.write('PE90');
   ods.write('PE95');
   ods.Write_MatR(SeriesPE_CEGH[5],'cmo3_PE_CEGH',2,1);
   end;




// 5.4  Fijo la cantidad de canales de cada corte de probabilidad
setlength(PE5_IMFIA,length(DatosEnsamble));
setlength(PE10_IMFIA,length(DatosEnsamble));
setlength(PE20_IMFIA,length(DatosEnsamble));
setlength(PE50_IMFIA,length(DatosEnsamble));
setlength(PE80_IMFIA,length(DatosEnsamble));
setlength(PE90_IMFIA,length(DatosEnsamble));
setlength(PE95_IMFIA,length(DatosEnsamble));
setlength(SeriesPE_IMFIA,length(DatosEnsamble));

for kSerie := 0 to (length(DatosEnsamble)-1) do
  begin
  // 5.5  Ordeno las series para sacar cada corte de probabilidad
  DatosEnsamble[kSerie].Sort();
  // 5.6  Cargo cada corte de porbabilidad a la serie correspondiente
  PE5_IMFIA[kSerie] := DatosEnsamble[kSerie].Columna(round(0.05*DatosEnsamble[kSerie].nc));
  PE10_IMFIA[kSerie]:=DatosEnsamble[kSerie].Columna(round(0.10*DatosEnsamble[kSerie].nc));
  PE20_IMFIA[kSerie]:=DatosEnsamble[kSerie].Columna(round(0.20*DatosEnsamble[kSerie].nc));
  PE50_IMFIA[kSerie]:=DatosEnsamble[kSerie].Columna(round(0.50*DatosEnsamble[kSerie].nc));
  PE80_IMFIA[kSerie]:=DatosEnsamble[kSerie].Columna(round(0.80*DatosEnsamble[kSerie].nc));
  PE90_IMFIA[kSerie]:=DatosEnsamble[kSerie].Columna(round(0.90*DatosEnsamble[kSerie].nc));
  PE95_IMFIA[kSerie]:=DatosEnsamble[kSerie].Columna(round(0.95*DatosEnsamble[kSerie].nc));
  // 5.7  Creo matrices con todos los cortes de probabilidad para cada canal
  SeriesPE_IMFIA[kSerie]:=TMatR.Create_Init(PE50_IMFIA[kSerie].n,7);
  SeriesPE_IMFIA[kSerie].pon_columna(1,1,PE5_IMFIA[kSerie].toTDAOfNReal());
  SeriesPE_IMFIA[kSerie].pon_columna(1,2,PE10_IMFIA[kSerie].toTDAOfNReal());
  SeriesPE_IMFIA[kSerie].pon_columna(1,3,PE20_IMFIA[kSerie].toTDAOfNReal());
  SeriesPE_IMFIA[kSerie].pon_columna(1,4,PE50_IMFIA[kSerie].toTDAOfNReal());
  SeriesPE_IMFIA[kSerie].pon_columna(1,5,PE80_IMFIA[kSerie].toTDAOfNReal());
  SeriesPE_IMFIA[kSerie].pon_columna(1,6,PE90_IMFIA[kSerie].toTDAOfNReal());
  SeriesPE_IMFIA[kSerie].pon_columna(1,7,PE95_IMFIA[kSerie].toTDAOfNReal());
  end;

// 5.8  Creo las hojas de los cortes de probabilidad
ods.agregoHoja('Bonete_PE_IMFIA');
if length(SeriesPE_IMFIA)>1 then
   ods.agregoHoja('Palmar_PE_IMFIA');
if length(SeriesPE_IMFIA)>2 then
   ods.agregoHoja('SG_PE_IMFIA');

// 5.9  Escribo el contenido de cada hoja
ods.go('Bonete_PE_IMFIA',1,1);
ods.write('PE5');
ods.write('PE10');
ods.write('PE20');
ods.write('PE50');
ods.write('PE80');
ods.write('PE90');
ods.write('PE95');
ods.Write_MatR(SeriesPE_IMFIA[0],'Bonete_PE_IMFIA',2,1);
if length(SeriesPE_IMFIA)>1 then
   begin
   ods.go('Palmar_PE_IMFIA',1,1);
   ods.write('PE5');
   ods.write('PE10');
   ods.write('PE20');
   ods.write('PE50');
   ods.write('PE80');
   ods.write('PE90');
   ods.write('PE95');
   ods.Write_MatR(SeriesPE_IMFIA[1],'Palmar_PE_IMFIA',2,1);
   end;
if length(SeriesPE_IMFIA)>2 then
   begin
   ods.go('SG_PE_IMFIA',1,1);
   ods.write('PE5');
   ods.write('PE10');
   ods.write('PE20');
   ods.write('PE50');
   ods.write('PE80');
   ods.write('PE90');
   ods.write('PE95');
   ods.Write_MatR(SeriesPE_IMFIA[2],'SG_PE_IMFIA',2,1);
   end;


///////////////////////////////////////////

// 5.10  Guardo el ods
ods.Guardar;
// 5.11  Libero la memoria usada
ods.free;
for kSerie := 0 to (length(DatosSimCEGH)-1) do
begin
  PE5_CEGH[kserie].free;
  PE10_CEGH[kserie].free;
  PE20_CEGH[kserie].free;
  PE50_CEGH[kserie].free;
  PE80_CEGH[kserie].free;
  PE90_CEGH[kserie].free;
  PE95_CEGH[kserie].free;
  SeriesPE_CEGH[kSerie].free;
end;
for kSerie := 0 to (length(DatosEnsamble)-1) do
begin
  PE5_IMFIA[kserie].free;
  PE10_IMFIA[kserie].free;
  PE20_IMFIA[kserie].free;
  PE50_IMFIA[kserie].free;
  PE80_IMFIA[kserie].free;
  PE90_IMFIA[kserie].free;
  PE95_IMFIA[kserie].free;
  SeriesPE_IMFIA[kSerie].free;
end;

SetLength(PE5_CEGH ,0);
SetLength(PE10_CEGH,0);
SetLength(PE20_CEGH,0);
SetLength(PE50_CEGH,0);
SetLength(PE80_CEGH,0);
SetLength(PE90_CEGH,0);
SetLength(PE95_CEGH,0);
SetLength(SeriesPE_CEGH,0);
SetLength(PE5_IMFIA,0);
SetLength(PE10_IMFIA,0);
SetLength(PE20_IMFIA,0);
SetLength(PE50_IMFIA,0);
SetLength(PE80_IMFIA,0);
SetLength(PE90_IMFIA,0);
SetLength(PE95_IMFIA,0);
SetLength(SeriesPE_IMFIA,0);


end;

procedure cargarEnsamblesLocales(var SeriesDeEnsambles: array of TMatR;
  archidir: string; Nseries: integer);
  // Carga los ramilletes de pronosticos del IMIFIA a partir de archivo .xlt
  // guardados localmente. Se usa para debuguear el programa.
begin
  SeriesDeEnsambles[0] := TMatR.CreateFromArchiXLT(archidir+'BonseriesIMFIA.xlt');
  if NSeries>1 then
     SeriesDeEnsambles[1] := TMatR.CreateFromArchiXLT(archidir+'PalseriesIMFIA.xlt');
  if NSeries>2 then
     SeriesDeEnsambles[2] := TMatR.CreateFromArchiXLT(archidir+'SGseriesIMFIA.xlt');

end;


procedure bajarEstadosInicialesHistIMFIA( var estado : TEstado_SIN_UY; dbpqPronos: TDBPQCon; Ndias: integer);
var
  diaInicioProno, kdia: integer;
  MatrizHist: TMatR;


begin
  MatrizHist := TMatR.Create_Init(Ndias,5);
  diaInicioProno := round(IsoStrToDateTime('2018-12-28'));
  for kdia:=1 to Ndias do
  begin
  writeln('Obteniendo ensambles de pronosticos de aportes. Dia '+ IntToStr(kdia) +' de '+ IntToStr(Ndias)+'.');
  estado.getPronoHidroIMFIA(dbpqPronos,diaInicioProno);

  MatrizHist.pon_e(kdia,1,diaInicioProno);
  MatrizHist.pon_e(kdia,2,estado.QBonIMFIAmatHora.Fila(1).promedio);
  MatrizHist.pon_e(kdia,3,estado.QBayIMFIAmatHora.Fila(1).promedio);
  MatrizHist.pon_e(kdia,4,estado.QPalIMFIAmatHora.Fila(1).promedio);
  MatrizHist.pon_e(kdia,5,estado.QSGIMFIAmatHora.Fila(1).promedio);
  diaInicioProno:=diaInicioProno+1;
  end;
  MatrizHist.WriteArchiXLT('Pronos_IMFIA_HIST.xlt');
  MatrizHist.free;
end;

procedure cargarEnsamblesDesdeEstado(var SeriesDeEnsambles: array of TMatR;
  estado: TEstado_SIN_UY; Nseries: integer);
  // Carga los ramilletes de pronosticos del IMFIA a partir del estado. En el
  // estado debe estar precargados los ramilletes a traves de consulta sql y
  // corrida del modelo del IMFIA.
begin
  SeriesDeEnsambles[0] := TMatR.Create_Clone(estado.QBonIMFIAmatHora);
  if NSeries>1 then
     SeriesDeEnsambles[1] := TMatR.Create_Clone(estado.QPalIMFIAmatHora);
  if NSeries>2 then
     SeriesDeEnsambles[2] := TMatR.Create_Clone(estado.QSGIMFIAmatHora);
end;

procedure graficarResultados_;
var
parametros : array of string;
begin
      // correr el script de octave
      {$IFDEF WINDOWS}
        RunChildAndWait('octave-cli', ['graficar_resultados_.m']);
      {$ELSE}
        setlength(parametros, 2);
        parametros[0] := '--no-window-system';
        //parametros[1] := '--no-gui';
        parametros[1] := '/home/' + GetEnvironmentVariable('USER') + '/prono_hidro_imfia/graficar_resultados.m';

        RunChildAndWait('/usr/bin/octave', parametros);
      {$ENDIF}
end;

procedure bajarHistoricos;
var
HistVates: TdatosHistoricosVates;
begin
     HistVates := TdatosHistoricosVates.Create;
     HistVates.descargarPronos;
     HistVates.armarIniciales;
     HistVates.free;
end;

procedure graficarResultados;
var
parametros : array of string;
begin
      // correr el script de octave
      {$IFDEF WINDOWS}
        RunChildAndWait('octave-cli', ['graficar_resultados.m']);
      {$ELSE}
        setlength(parametros, 1);
        parametros[0] := '/home/' + GetEnvironmentVariable('USER') + '/prono_hidro_imfia/graficar_resultados.m';

        RunChildAndWait('/usr/bin/octave', parametros);
      {$ENDIF}
end;

{ TdatosHistoricosVates }

constructor TdatosHistoricosVates.Create;

var
  kSerie: Integer;

begin
series:=4;

NombresSeries[0] := 'qapbon';
NombresSeries[1] := 'qapbay';
NombresSeries[2] := 'qappal';
NombresSeries[3] := 'qapsg';
NombresSeries[4] := 'pronosaportessaltooficial';

FechaIni := trunc(IsoStrToDateTime('2018-01-01 00:00:00') - 2);
FechaFin := trunc(IsoStrToDateTime('2019-03-27 00:00:00'));
Fecha    := trunc(FechaFin - 2);
dias     := trunc(Fecha - FechaIni);

horas    := 24*17;

EstadosIniciales := TSeriesDeDatos.CreateVacia(FechaIni,24,dias,1);

sqlbase1 := 'SELECT dt_cotas,dt_aportes';


for kSerie:=0 to series-1 do
    begin
    sqlbase1 := sqlbase1 + ',' + NombresSeries[kSerie] ;
    end;

EstadosIniciales := TSeriesDeDatos.CreateVacia(FechaIni,24,dias,1);

sqlbase1 := sqlbase1 + ' FROM sinuy_estado_y_pronos WHERE dt_aportes <= ''';
sqlbase2 := ''' ORDER BY dt_cotas DESC,dt_aportes LIMIT 1 ';

setlength(pronosticos,dias);
setlength(pronosticosPasoDiario,dias);

end;

procedure TdatosHistoricosVates.descargarPronos;


var
  kDia, kSerie: Integer;
  sql, Strdt_aportes, ApStr: String;
  r:TDataRecord;
  ApVector: TVectR;
  ds: urosx.TSQLQuery;

begin

for kDia := 0 to dias-1 do
begin
    sql := sqlbase1 + DateTimeToIsoStr(Fecha) + sqlbase2;
    writeln(sql);
    ds:=dbcon_rosx.query(sql);
    r:=ds.first;

    Strdt_aportes := r.GetByNameAsString('dt_aportes');
    pronosticos[kdia] := TSeriesDeDatos.CreateVacia(Fecha ,1,horas,1);
    for kSerie := 0 to series-1 do
    begin
         ApStr := r.GetByNameAsString(NombresSeries[kSerie]);
         ApVector := TVectR.Create_unserialize_pg(ApStr);
         if ApVector.n < horas then
            ApVector.resize(horas);

         pronosticos[kDia].AddSerie(NombresSeries[kSerie],ApVector);
    end;
    pronosticos[kdia].WriteToArchi('/home/'+ GetEnvironmentVariable('USER')+ '/PronosticosVates/pronosticosHorarios-' + DateTimeToIsoStr(Fecha) + '.xlt');
    Fecha:= Fecha-1;
end;



end;

procedure TdatosHistoricosVates.armarIniciales;

var
  kdia,kSerie: integer;
  VectorEstadoInicial: array of TVectR;
begin

//puntos:=16;

setlength(VectorEstadoInicial,series);

for kserie:=0 to series-1 do
    VectorEstadoInicial[kserie]:= TVectR.Create_InitVal(dias,0);

for kdia:=0 to dias-1 do
begin
     pronosticosPasoDiario[kdia] := pronosticos[kdia].Resampling(24);
     for kserie:=0 to series-1 do
         VectorEstadoInicial[kserie].pon_e(dias-kdia,pronosticosPasoDiario[kdia].GetSerie(kserie).e(3));

end;


for kserie:=0 to series-1 do
    EstadosIniciales.AddSerie(NombresSeries[kSerie],VectorEstadoInicial[kserie]);

EstadosIniciales.WriteToArchi('/home/'+GetEnvironmentVariable('USER')+ '/PronosticosVates/Guias/EstadosInicialesHistVates.xlt');


for kserie:=0 to series-1 do
    VectorEstadoInicial[kserie].free;
setlength(VectorEstadoInicial,0);

end;

procedure TdatosHistoricosVates.free;

var
  kserie: Integer;
begin

if EstadosIniciales <> nil then
     EstadosIniciales.free;

if pronosticos <> nil then
  for kserie:=0 to High(pronosticos) do
      if pronosticos[kserie]<>nil then
         pronosticos[kserie].free;


if pronosticosPasoDiario <> nil then
 for kserie:=0 to high(pronosticosPasoDiario) do
     if pronosticosPasoDiario[kserie]<>nil then
        pronosticosPasoDiario[kserie].free;


end;


{ TProblemaResolverFactores }

constructor TProblemaResolverFactores.Create(aM: TMatR);
var
  hcanal: Integer;
begin
// 3.4.1  Creo el problema como hijo de la clase TProblema_m01
inherited Create_init(1, aM.nc + 1, 0, nil, nil);

// 3.4.2  Cargo la matriz M y Ncanales para tenerlos disponibles
M := aM;
Ncanales := M.nc;

// 3.4.3  Creo la función objetivo en base a la matriz M
f := T_fobjetivoCEGH.Create(M);

// 3.4.4  Cargo el valor inicial de B
Betas:= TVectR.Create_InitVal(Ncanales,1e+4);

// 3.4.5  Fijo las restricciones de caja
for hcanal := 1 to Ncanales do
    begin
    cota_inf_set(hcanal, 1e-40);
    cota_sup_set(hcanal, 1e+40);
    end;

end;

procedure TProblemaResolverFactores.free;
begin
//  Betas.free;
  inherited free;
end;

function TProblemaResolverFactores.resolverFactores(var factores:TMatR): boolean;
var
X : TVectR;
l : TVectR;
Convergio: boolean;
hcanal: integer;
cont_Iters: longint;
fCosto, fLagrange, dFrontera2: NReal;
begin

// 3.5.1
// Creo un vector de restricciones lambda (l) vacio
l := TVectR.Create_init(1);
l.pon_e(1, 0);

// 3.5.2
// Creo el vector X de resolucion del problema. Lo inicio en el valor de Beta.
// En el primer paso tiene un valor precargado, para el resto utiliza como punto
// de partida el resultado en el paso anterior
X := TVectR.Create_Init(Ncanales);
for hcanal:= 1 to Ncanales do
begin
     X.pon_e(hcanal,Betas.e(hcanal));
end;

// 3.5.3
// Para la resolucion uso un MaxInBox_Dual, con restricciones vacias.
MaxInBox_Dual(l,X,1000,0.01,100000,cont_Iters,fCosto,fLagrange, dFrontera2, convergio,false);

// 3.5.4
// Cargo en Betas el resultado de la optimizacion
Betas := TVectR.Create_Clone(X);

// 3.5.5
// Cargo los factores: 1/Beta
for hcanal:=1 to Ncanales do
begin
     //factores.pon_e(hcanal,1,1/Betas.e(hcanal));
     factores.pon_e(hcanal,1,1/Betas.e(hcanal));
end;

// 3.5.6
// Libero la memoria usada
l.free;
X.free;
result:= convergio;

end;





{ T_fobjetivoCEGH }

constructor T_fobjetivoCEGH.Create(aM: TMatR);
// 3.4.3
begin
// 3.4.3.0 Creo el T_fobjetivoCEGH en base al padre
     inherited Create;

// 3.4.3.1  Cargo la matriz del problema
self.M := aM;
self.Ncanales := M.nc;

// 3.4.3.2  Construyo FX que es el vector FX = M * X -"1/X", que buscamos minimizar.
self.FX  := TVectR.Create_Init(Ncanales);

M.WriteArchiXLT('/home/gflieller/prono_hidro_imfia/M.xlt');

end;

function T_fobjetivoCEGH.f(const X: TVectR): NReal;
// Calculo del error
var
X_1: TVectR;
hcanal: Integer;
error : NReal;
begin
// Esto se usa adentro del MaxInBox que es el 3.5.3
// 3.5.3.1  Creo X_1 que es lo que tambien se llamo "1/X"
X_1 := TVectR.Create_Init(Ncanales);
for hcanal:=1 to Ncanales do
    begin
    X_1.pon_e(hcanal,-1/X.e(hcanal));
    end;

// 3.5.3.2  Calculo M * X - "1/X" que es la funcion de error
FX.Mult(M,X);                        // FX = M * X
FX.Suma(FX,X_1);                     // FX = M * X - "1/X"

// 3.5.3.3  Se da como resultado la norma euclidea por un factor amplificador
//          para poder manejar el error evitando errores numericos.
error:= FX.normEuclid;         // error = 1e+08*RAIZ(FX * FX^T)

X_1.free;
result := error;

end;

procedure T_fobjetivoCEGH.free;
begin
  FX.free;
end;

procedure T_fobjetivoCEGH.acum_g(var grad: TVectR; const X: TVectR);
// Calculo del gradiente del error
var
  kderivada, hcanal: Integer;
  derivada: NReal;
begin
// 3.5.3.4
// error = 1e+08*RAIZ(FX * FX^T) = 1e+08*RAIZ( suma(f_j(x)^2) )
// derror/dx_i = ( 1e+08/(2*RAIZ(error)) ) * suma( 2*f_j(x)*df_j/dx_i))
// derror/dx_i = ( 1e+08/(2*RAIZ(error)) ) * suma( 2*f_j(x)*(
for kderivada:=1 to Ncanales do
    begin
    derivada := 0;
    if FX.normEuclid <> 0 then
    begin
        for hcanal:=1 to Ncanales do
             derivada := derivada + 2*FX.e(hcanal)*M.e(hcanal,kderivada);

        derivada := derivada + 2*FX.e(kderivada)*(1/power(X.e(kderivada),2));

        derivada := (derivada/(2*FX.normEuclid));
    end;
    grad.acum_e(kderivada,derivada);

    end;

end;

{ TopcionesCalibrarConoEnsamble }


{ TcalibrarConoEnsamble }

constructor TcalibrarConoEnsamble.Create(acegh: TFuenteSintetizadorCEGH; aEnsambles: array of TMatR);

begin
inherited Create;

self.fcegh := acegh;

cargarDimensiones(aEnsambles);

// 2.1 C es la misma matriz que se menciona en la documentación
cargarFiltros;

// 2.2 Pasa ensamble a Gaussiano
cargarEnsamblesGausiano(aEnsambles);

// 2.3 Calcula y carga las guias
calcularGuias;

// 2.4 Calcula y carga los desvios respecto a la guia
calcularDesvios;

// 2.5 Calculo la matriz de covarianzas
// Para cada miembro del ensamble y para cada paso se calcula des_desT := des * des^T. desT = des^T
calcularCovar;

// 2.6 Calcula los sesgos
calcularSesgos;


end;

function TcalibrarConoEnsamble.calibrarConoEnsamble(): boolean;

var

Z0, Z, LambdaAux: TMatR;
kpaso : integer;
convergio: boolean;
Problema: TProblemaResolverFactores;
{$IFDEF MODO_DEBUG_ENSAMBLES}
ZetasBon: TVectR;
{$ENDIF}
begin
// 3.1 Inicializo
Setlength(betas, Npasos);
setlength(Lambdas,Npasos);
setlength(M,Npasos);
Z0 := TMatR.Create_Init(Ncanales,Ncanales);
{$IFDEF MODO_DEBUG_ENSAMBLES}
ZetasBon := TVectR.Create_Init(Npasos);
{$ENDIF}
for kPaso := 0 to Npasos-2 do
  begin
  // 3.2 Calculo la matriz de ajuste del paso anterior Z
  Z := calcularZ(Z0,kPaso);
  // 3.3 Calculo M en base a la covarianza y Z
  armarProblema(Z, kPaso);
  // 3.4 Creo el problema calculo de Beta en base a la matriz M
  Problema := TProblemaResolverfactores.Create(M[kpaso]);
  // 3.5 Resuelvo M * X = "1/X"
  LambdaAux := TMatR.Create_Init(Ncanales,1);
  convergio := Problema.resolverFactores(LambdaAux);
  Lambdas[kpaso] := TMatR.Create_Clone(LambdaAux);
  LambdaAux.free;
  Problema.free;
  // 3.6 Preservo Z para el paso siguiente como informacion del paso previo
  Z0 := Z;
  end;
Z0.free;
// 3.7. Cargo los factores lambda a un ods
ImprimirLambdas;

// 3.8 Convierte los resultados al formato que usa el CEGH
ArmarResultados(fcegh.datosModelo_Sim);

result := convergio;

end;


procedure TcalibrarConoEnsamble.cargarDimensiones(aEnsambles: array of TMatR);
begin
Ncanales := length(aEnsambles);
Nmiembros := aEnsambles[0].nc;
Npasos := aEnsambles[0].nf;
end;

procedure TcalibrarConoEnsamble.cargarEnsamblesGausiano(aEnsambles: array of TMatR);
var
hcanal,jmiembro,kpaso : integer;
ods: TLibroOpenCalc;
begin

// 2.2.1 Paso cada ensamble a mundo gaussiano, en forma de array de matrices.
{$IFDEF GFTEST}
// Si estoy  en el modo de prueba voy guardando las series y las matrices en un ods
ods := TLibroOpenCalc.Create(TRUE,'Series_Gaussianas.ods',False,TsSpreadsheetFormat.sfOpenDocument);
{$ENDIF}
setlength(EnsamblesGauss, length(aEnsambles));
for hcanal := 1 to Ncanales do
    begin
        EnsamblesGauss[hcanal-1] := TransformarEnsamble(fcegh.datosModelo_Sim, aEnsambles, hcanal);

        {$IFDEF GFTEST}
        EnsamblesGauss[hcanal-1].WriteArchiXLT('SerieGaussiana_'+IntToStr(hcanal-1)+'.xlt');
        ods.agregoHoja('SerieGaussiana_'+IntToStr(hcanal-1));
        ods.Write_MatR(EnsamblesGauss[hcanal-1],'SerieGaussiana_'+IntToStr(hcanal-1),1,1);
        {$ENDIF}
    end;
{$IFDEF GFTEST}
ods.agregoHoja('Matriz_A');
ods.Write_MatR(self.A,'Matriz_A',1,1);
ods.agregoHoja('Matriz_B');
ods.Write_MatR(fcegh.datosModelo_Sim.B_cte,'Matriz_B',1,1);
ods.agregoHoja('Matriz_C');
ods.Write_MatR(self.C,'Matriz_C',1,1);
ods.Guardar;
ods.free;
{$ENDIF}


// 2.2.2 Convierto los arrays de matrices en arrays de array de matrices columna
//       para poder realizar multiplicaciones mas facilmente
// Tengo m (M = Ncanales) matrices de r*p (r=Nmiembros,p=Npasos). Los vectores
// X_k deben ser de m*1 y ser p vectores, 1 por paso, para los r miembros.
// X[miembro][paso] = (X_jkc1, X_jkc2,..., X_jkcm)
setlength(Xg, Nmiembros);
for jmiembro := 1 to Nmiembros do
begin
    setlength(Xg[jmiembro-1],Npasos);
    for kpaso := 1 to Npasos do
    begin
        Xg[jmiembro-1][kpaso-1] := TMatR.Create_InitVal(Ncanales,1,0);
        for hcanal := 1 to Ncanales do
        begin
            Xg[jmiembro-1][kpaso-1].pon_e( hcanal,1, EnsamblesGauss[hcanal-1].e(kpaso,jmiembro) );
        end;
    end;
end;

end;

procedure TcalibrarConoEnsamble.calcularGuias;

var
guias: array of TDAofNReal;
guiasgaussianas : array of TDAofNReal;
hcanal, kpaso : integer;
matGuiasGuass: TMatR;
begin

// 2.3.1
// Creo el pronostico en base a los datos del modelo de simulacion del CEGH
aPronosticos:= Tpronosticos.Create_Default(fcegh.datosModelo_Sim);
setlength(aPronostico,Ncanales);
setlength(guias,Ncanales);
setlength(guiasGaussianas,Ncanales);

// 2.3.2 creo y cargo las guias gaussianas y reales y las cargo al pronostico
for hcanal:=0 to (Ncanales-1) do
begin

  setlength(guiasGaussianas[hcanal],Npasos);
  setlength(guias[hcanal],Npasos);
  // Armo la guia en el mundo gaussiano y paso su transformada al mundo real
  for kpaso:=0 to Npasos-1 do
  begin
      guiasGaussianas[hcanal][kpaso] := EnsamblesGauss[hcanal].Fila(kpaso+1).promedio;
      guias[hcanal][kpaso] := fcegh.datosModelo_Sim.gTox(guiasGaussianas[hcanal][kpaso],hcanal+1,fcegh.escenariosDePronosticos[0].fechaIniCono.horasDesdeElInicioDelAnio+kpaso);
  end;

  // Cargo la guia en el aPronosticos de la clase TcalibrarConoEnsamble
  aPronostico[hcanal] := TPronostico.Create_Default(fcegh.datosModelo_Sim);
  aPronostico[hcanal].guia_pe:=0.5;
  aPronostico[hcanal].guia_eg:=guiasgaussianas[hcanal];
  aPronostico[hcanal].guia:=guias[hcanal];

end;

// 2.3.3 Calculo Xesp = <X_k>, p vectores de m*1 entradas, promedios por canal de miembros
setlength(Xesp,Npasos);
for kpaso := 0 to Npasos-1 do
    begin
    Xesp[kpaso]:= TMatR.Create_Init(Ncanales,1);
    for hcanal := 0 to Ncanales-1 do
        Xesp[kpaso].pon_e(hcanal+1,1,guiasgaussianas[hcanal][kpaso]);
end;

{matGuiasGuass:= TMatR.Create_Init(Ncanales,Npasos);
for kpaso := 0 to Npasos-1 do
    for hcanal:=0 to Ncanales-1 do
        matGuiasGuass.pon_e(hcanal+1,kpaso+1,Xesp[kpaso].e(hcanal+1,1));

matGuiasGuass.WriteArchiXLT('Guia_Gaussiana.xlt');}

end;


procedure TcalibrarConoEnsamble.calcularSesgos;
// Calculo los sesgos
var
kpaso: integer;
Saux: TMatR;
sesgoVect:TVectR;

begin
// 2.6
setlength(S,Npasos-1);
for kpaso := 1 to Npasos-1 do
begin
     Saux:= TMatR.Create_Clone(Xesp[kPaso-1]);
     Saux.Mult(A,Saux);                                             // A*X0
     S[kpaso-1]:= TMatR.Create_Combinar(1,Xesp[kpaso],-1,Saux);     // S = X1 - A*X0
     Saux.free;
end;
sesgoVect:=TVectR.Create_Init(Npasos-1);
for kpaso:=1 to Npasos-1 do
    sesgoVect.pon_e(kpaso,S[kpaso-1].e(3,1));
sesgoVect.WriteXLTSimple_archi('/home/gflieller/prono_hidro_imfia/sesgoSG.xlt');
// Ahora S es un array de matrices (uso matrices para poder hacer multiplicacion)
end;


procedure TcalibrarConoEnsamble.calcularCovar;
// 2.5.
// Calculo las matrices covarianzas
// p matrices de covarianza de dimesion m x m, promedio de las matrices de des * des^T
// En la documentación se le llamo Xi
var
jmiembro,kpaso : integer;
des_desT, desT: array of array of TMatR;
CovarVector: TVectR;

begin
// 2.5.1 Calculo d*d^T para todos los pasos
setlength(desT, Nmiembros);
setlength(des_desT, Nmiembros);
for jmiembro := 0 to (Nmiembros-1) do
begin
    setlength(desT[jmiembro],Npasos);
    setlength(des_desT[jmiembro],Npasos);
    for kpaso := 0 to (Npasos-1) do
    begin
         desT[jmiembro][kpaso] := TMatR.Create_Clone(des[jmiembro][kpaso]);
         desT[jmiembro][kpaso].Transponer;
         des_desT[jmiembro][kpaso] := TMatR.Create_InitVal(Ncanales,Ncanales,0);
         des_desT[jmiembro][kpaso].Mult(des[jmiembro][kpaso],desT[jmiembro][kpaso]);
    end;
end;

// 2.5.2 Calculo las matrices de covarianza para todos los pasos
setlength(covar, Npasos);
for kpaso := 0 to (Npasos-1) do
begin
     covar[kpaso] := TMatR.Create_InitVal(Ncanales,Ncanales,0);
     for jmiembro := 0 to Nmiembros-1 do
          covar[kpaso].Suma(covar[kpaso],des_desT[jmiembro][kpaso]);

     covar[kpaso].PorReal(1/(Nmiembros-1));
//     covar[kpaso].WriteArchiXLT('covar'+IntToStr(kpaso)+'.xlt');
end;  // Debug: Esta igual que en el metodo B -- GF@2019-02-12 11:58


// 2.5.3 Libero la memoria que utilice
for jmiembro:=0 to (Nmiembros-1) do
begin
    for kpaso:=0 to (Npasos-1) do
    begin
        desT[jmiembro][kpaso].Free;
        des_desT[jmiembro][kpaso].Free;
    end;
    SetLength(desT[jmiembro],0);
    SetLength(des_desT[jmiembro],0);
end;
SetLength(desT,0);
SetLength(des_desT,0);

end;


procedure TcalibrarConoEnsamble.calcularDesvios;
var
X_normal_0 : array of array of TMatR;
X_normal_0_promed : array of TMatR;
p : NReal;
jmiembro, kpaso, hcanal: Integer;


// 2.4.1 Normalizo los valores en el mundo gaussiano con respecto al promedio
begin
setlength(X_normal_0, Ncanales);
for hcanal:=0 to Ncanales-1 do
begin
SetLength(X_normal_0[hcanal],Npasos);
    for kpaso:=0 to Npasos-1 do
    begin
        X_normal_0[hcanal][kpaso] := TMatR.Create_Init(Nmiembros,1);
        for jmiembro := 0 to Nmiembros-1 do
              X_normal_0[hcanal][kpaso].pon_e(jmiembro+1,1,  Xg[jmiembro][kpaso].e(hcanal+1,1)-Xesp[kpaso].e(hcanal+1,1)  );
              //X_normal_0[hcanal][kpaso].pon_e(jmiembro+1,1,  Xg[jmiembro][kpaso].e(hcanal+1,1)  );
        X_normal_0[hcanal][kpaso].Sort();
    end;
end;    // Esto esta igual que en el metodo b. (X-guia) GF@2019-02-11-11:46




// 2.4.2 Calculo el promedio por paso de la X normalizada
SetLength(X_normal_0_promed,Npasos);
for kpaso := 0 to Npasos-1 do
begin
X_normal_0_promed[kpaso] := TMatR.Create_Init(Ncanales,1);
    for hcanal := 0 to Ncanales-1 do
      begin
      p:=0;
           for jmiembro:=0 to Nmiembros-1 do
               p:= p + X_normal_0[hcanal][kpaso].e(jmiembro+1,1);
      p:=p/Nmiembros;
      X_normal_0_promed[kpaso].pon_e(hcanal+1,1, p )  ;
      end;
end;


// 2.4.3 Cargo las desviaciones
SetLength(des,Nmiembros);
for jmiembro := 0 to Nmiembros-1 do
    begin
       SetLength(des[jmiembro],Npasos);
       For kpaso := 0 to Npasos-1 do
       begin
          des[jmiembro][kpaso]:= TMatR.Create_Init(Ncanales,1);
          for hcanal:=1 to Ncanales do
              des[jmiembro][kpaso].pon_e( hcanal,1, X_normal_0[hcanal-1][kpaso].e(jmiembro+1,1) - X_normal_0_promed[kpaso].e(hcanal,1) );
              //des[jmiembro][kpaso].pon_e( hcanal,1, X_normal_0[hcanal-1][kpaso].e(jmiembro+1,1) );
       end;
    end;
{$IFDEF MODO_DEBUG_ENSAMBLES}

{$ENDIF}



// 2.4.4  Libero memoria
for kpaso := 0 to Npasos-1 do
begin
   X_normal_0_promed[kpaso].free;
end;
SetLength(X_normal_0_promed,0);

for hcanal:=0 to Ncanales-1 do
begin
    for kpaso:=0 to Npasos-1 do
    begin
         X_normal_0[hcanal][kpaso].Free;
    end;
    SetLength(X_normal_0[hcanal],0);
end;
SetLength(X_normal_0,0);


end;

function TcalibrarConoEnsamble.calcularZ(Z0: TMatR; kpaso: integer) : TMatR;
var
Z, AT, BT, L, Zaux:TMatR;
hcanal: Integer;
begin
  if kpaso = 0 then
     // 3.2.1 El Z0 se calcula como Z0 = A * d * d^T * A^T
     begin
     Z := TMatR.Create_Clone(A);
     Z.Transponer;
     Z.Mult(covar[kpaso],Z);
     Z.Mult(A,Z);
     end
  else
     begin
     // 3.2.2 Z = A *( Z0 + L * B * B^T * L^T )* A^T
     // 3.2.2.1 Calculo las traspuestas
     AT := TMatR.Create_Clone(A);
     AT.Transponer;
     BT := TMatR.Create_Clone(B);
     BT.Transponer;

     // 3.2.2.2 Calculo la matriz L, la cual es diagonal. L^T = L
     L := TMatR.Create_InitVal(Ncanales,Ncanales,0);
     for hcanal := 1 to Ncanales do
         L.pon_e(hcanal,hcanal,Lambdas[kpaso-1].e(hcanal,1));

     // 3.2.2.3 Calculo Zaux = L * B * B^T * L^T
     Zaux:= TMatR.Create_Clone(Z0);
     Zaux.Mult(BT,L);
     Zaux.Mult(B,Zaux);
     Zaux.Mult(L,Zaux);

     // 3.2.2.4 Calculo Zaux = Z0 + L * B * B^T * L^T
     Zaux.Suma(Z0,Zaux);

     // 3.2.2.4 Calculo Z = A *( Z0 + L * B * B^T * L^T )*A^T
     Zaux.Mult(Zaux,AT);
     Z := TMatR.Create_InitVal(Ncanales,Ncanales,0);
     Z.Mult(A,Zaux);

     // 3.2.2.5 Libero la memoria que use
     AT.Free;
     BT.Free;
     Zaux.Free;
     L.free;
     end;

Z.WriteArchiXLT('Z.xlt');
result := Z;

end;


procedure TcalibrarConoEnsamble.cargarFiltros;
// Cargo las matrices A y B. Calculo C = (B * B^T)^-1.
begin
  A := fcegh.datosModelo_Sim.A_cte;
  B := fcegh.datosModelo_Sim.B_cte;
  C := TMatR.create_clone(B);
  C.Transponer;
  C.Mult(B,C);
  C.Inv;

//  A.WriteArchiXLT('A.xlt');
//  B.WriteArchiXLT('B.xlt');
//  C.WriteArchiXLT('C.xlt');
end;


procedure TcalibrarConoEnsamble.armarProblema(Z : TMatR; kpaso : integer);
// 3.3 Calculo la matriz M del problema

var
  Cinv: TMatR;
begin
if kpaso=0 then
   // 3.3.1 M0 = d*d^T . C
   begin
   M[kpaso] := TMatR.Create_Clone(covar[kpaso]);
   M[kpaso].MaximizarElementoAElemento(0);
   Cinv := TMatR.Create_Clone(C);
   Cinv.Inv;
   //M[kpaso].porMatrizElementoAElemento(Cinv);
   M[kpaso].porMatrizElementoAElemento(C);
   Cinv.free;
   end
else
    // 3.3.2 M = (d*d^T - Z). C
    begin
    M[kpaso] := TMatR.Create_Combinar(1,covar[kpaso],-1,Z);
    M[kpaso].MaximizarElementoAElemento(0);
    M[kpaso].porMatrizElementoAElemento(C);

    end;

end;




procedure TcalibrarConoEnsamble.ImprimirLambdas;
// Guarda en un archivo xlt los valores de los factores lambda

var
Lambda : TMatR;
hcanal, kpaso : integer;
fechaAhora: TFecha;
begin

// 3.7.1  Creo una matriz para conter en cada columna los lambda de cada canal
//       de cada paso
Lambda:=TMatR.Create_Init(Ncanales,Npasos);
for kpaso:=1 to Npasos-1 do
    for hcanal:=1 to Ncanales do
    begin
        Lambda.pon_e(hcanal,kpaso, Lambdas[kpaso-1].e(hcanal,1));
    end;

// 3.7.2  Guardo la matriz en xlt
fechaAhora:= TFecha.Create_Dt(Now);
Lambda.WriteArchiXLT('/home/' + GetEnvironmentVariable('USER') + '/prono_hidro_imfia/Resultados/Lambdas/Lambdas_'+fechaAhora.AsAAAAMMDDhhmmtr+'.xlt');

//  3.7.3  Libero la memoria usada
Lambda.free;
fechaAhora.free;
end;


function TcalibrarConoEnsamble.TransformarEnsamble(aModelo:TModeloCEGH ;aEnsambles: array of TMatR; hcanal: integer): TMatR;
var
jmiembro,kpaso,kPasoHorarioDelAnio:integer;
MatAux:TMatR;
begin
MatAux:=TMatR.Create_Init(Npasos,Nmiembros);

writeln('Paso inicial: '+ IntToSTr(fcegh.escenariosDePronosticos[0].fechaIniCono.horasDesdeElInicioDelAnio));
     for jmiembro := 1 to Nmiembros do
     begin
         for kpaso := 1 to Npasos do
         begin
         kPasoHorarioDelAnio := fcegh.escenariosDePronosticos[0].fechaIniCono.horasDesdeElInicioDelAnio+kpaso;
         if (kPasoHorarioDelAnio>24*364) then
            kPasoHorarioDelAnio := kPasoHorarioDelAnio - 24*364;
         MatAux.pon_e(kpaso,jmiembro,  aModelo.xTog(aEnsambles[hcanal-1].e(kpaso,jmiembro),hcanal,kPasoHorarioDelAnio) );
         end;

     end;
{$IFDEF MODO_DEBUG_ENSAMBLES}
MatAux.WriteArchiXLT('Ensamble_Gaussiano.xlt');
{$ENDIF}
result := TMatR.Create_Clone(MatAux);
MatAux.free;
end;


procedure TcalibrarConoEnsamble.ArmarResultados(aModelo: TModeloCEGH);
//  Pasa los resultados de la calibracion a
//  los formatos en que los utiliza el CEGH
var
hcanal, kpaso , i: integer;
// 3.8   Los sesgos, guias y factores tienen que ir cargados en Pronostico como
//       TDAofNReal. Aca nos ocupamos de cargar unicamente los valores de los
//       sesgos y factores, las guias se cargan aparte.
begin
// 3.8.1  Para cada canal se crea un TDAofNReal.
setlength(sesgos, Ncanales);
setlength(factores,Ncanales);

for hcanal:=0 to (Ncanales-1) do
begin
// 3.8.2  Cada TDofNReal tiene como largo la cantidad de pasos del pronostico
  setlength(sesgos[hcanal],Npasos);
  setlength(factores[hcanal],Npasos);
  for kpaso:=0 to Npasos-2 do
  begin
      sesgos[hcanal][kpaso] := S[kpaso].e(hcanal+1,1);
      factores[hcanal][kpaso] := Lambdas[kpaso].e(hcanal+1,1);
  end;
end;

// 3.8.3  Cargo al Pronostico los parametros del cono. NPSA = -111111 indica que
//        los parametros no se deben considerar. Lo cargo en NPSA porque los
//        otros parametros pueden usarse para otras cosas
for hcanal := 0 to Ncanales-1 do
begin
aPronostico[hcanal].NPCC := 336;
aPronostico[hcanal].NPSA := -1;
aPronostico[hcanal].NPAC := 336;

// 3.8.4  Cargo los sesgos y factores a cada Pronostico
setlength(aPronostico[hcanal].factor, 336);
vclear(aPronostico[hcanal].factor);
aPronostico[hcanal].factor:= factores[hcanal];
vclear(aPronostico[hcanal].sesgo);
aPronostico[hcanal].sesgo := sesgos[hcanal];

// 3.8.5  Cargo cada Pronostico al Pronosticos
aPronosticos.setItem(hcanal,aPronostico[hcanal]);
end;
aPronosticos.Probabilidad:=1;
aPronosticos.fechaIniCono.dt := fcegh.escenariosDePronosticos[0].fechaIniCono.dt;
aPronosticos.prepararse(1);

// 3.8.6  Libero memoria innecesaria
for kpaso:=0 to Npasos-2 do
begin
     S[kpaso].free;
     Lambdas[kpaso].free;
end;
setlength(S,0);
setlength(Lambdas,0);

end;

procedure TcalibrarConoEnsamble.free;
var
k, j:integer;
begin
  if A<>nil then
     A.free;
  if B<>nil then
     B.free;
  if C<>nil then
     C.free;
  if aPronosticos<> nil then
     aPronosticos.Free;
  if fcegh<> nil then
     fcegh.Free;
  if Beta<>nil then
     Beta.free;
  if Lambda<>nil then
       Lambda.free;

  if length(Z)<>0 then
     begin
     for k:=0 to High(Z) do
         if Z[k] <> nil then
            Z[k].Free;
     SetLength(Z,0);
     end;

  if length(M)<>0 then
     begin
     for k:=0 to High(M) do
         if M[k] <> nil then
            M[k].Free;
     SetLength(M,0);
     end;

  if length(S) <> 0 then
    begin
    for k:=0 to High(S) do
         if S[k] <> nil then
            S[k].Free;
    SetLength(S,0);
    end;

  if length(EnsamblesGauss) <> 0 then
      begin
      for k:=0 to High(EnsamblesGauss) do
           if EnsamblesGauss[k] <> nil then
              EnsamblesGauss[k].Free;
      SetLength(EnsamblesGauss,0);
      end;

  if length(Xesp) <> 0 then
      begin
      for k:=0 to High(Xesp) do
           if Xesp[k] <> nil then
              Xesp[k].Free;
      SetLength(Xesp,0);
      end;

  if length(betas) <> 0 then
     begin
     for k:=0 to High(betas) do
         if betas[k] <> nil then
            betas[k].free;
     SetLength(betas,0);
     end;

  if length(Lambdas) <> 0 then
     begin
     for k:=0 to High(Lambdas) do
         if Lambdas[k] <> nil then
            Lambdas[k].free;
     SetLength(Lambdas,0);
     end;

  if length(covar) <> 0 then
     begin
     for k:=0 to High(covar) do
         if covar[k] <> nil then
            covar[k].free;
     SetLength(covar,0);
     end;

  if length(Xg) <> 0 then
      begin
      for k:=0 to High(Xg) do
         if length(Xg[k]) <> 0 then
         begin
              for j:=0 to High(Xg[k]) do
              if Xg[k][j] <> nil then
                 Xg[k][j].Free;
              setlength(Xg[k],0);
         end;
      SetLength(Xg,0);
      end;

  if length(des) <> 0 then
      begin
      for k:=0 to High(des) do
         if length(des[k]) <> 0 then
         begin
              for j:=0 to High(des[k]) do
              if des[k][j] <> nil then
                 des[k][j].Free;
              setlength(des[k],0);
         end;
      SetLength(Xg,0);
      end;

end;

procedure TcalibrarConoEnsamble.guardarSala(sala: TSalaDeJuego);
// Guarda la sala con los cambios que se hicieron durante la iteracion actual

var
  f: TArchiTexto;
  dirTemporales: array [0 .. MAX_PATH] of char;
  archi: String;

begin

  dirTemporales := GetTempDir;

  f := TArchiTexto.CreateForWrite(dirTemporales + 'guardar_temp.ese', True, 1);
  try
    begin
      f.wr('sala', TSalaDeJuego(sala));
      f.Free;

      archi := sala.archisala_;
      f := TArchiTexto.CreateForWrite(archi, True, 1);
      f.wr('sala', TSalaDeJuego(sala));

      sala.setDirCorrida(archi);
    end
  finally
    f.Free;
  end;
end;





end.

